import json
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Type, TypeVar

import torch
from torch import Tensor, nn
from torch.nn.utils.rnn import PackedSequence
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    PreTrainedModel,
    AutoModelForTokenClassification,
)

from kilroy_module_huggingface.utils import suppress_huggingface_logging
from kilroy_module_pytorch_py_sdk import (
    SequentialModel,
    Savable,
    background,
    pack_padded,
    unpack_to_padded,
)

BaseType = TypeVar("BaseType", bound="HuggingfaceSequentialModelBase")


def make_mask(x: Tensor, lengths: Tensor) -> Tensor:
    indices = torch.arange(x.shape[-1]).repeat((len(x), 1))
    lengths = lengths.view(-1, 1)
    return indices < lengths


class HuggingfaceSequentialModelBase(SequentialModel, Savable, ABC):
    def __init__(self, model: PreTrainedModel, pad_token_id: int) -> None:
        super().__init__()
        self._model = model
        self._pad_token_id = pad_token_id

    @classmethod
    @abstractmethod
    async def from_path(cls: Type[BaseType], path: str) -> BaseType:
        pass

    async def save(self, directory: Path) -> None:
        model_directory = directory / "model"
        config_path = directory / "config.json"
        model_directory.mkdir(parents=True, exist_ok=True)

        await background(self._model.save_pretrained, model_directory)
        with open(config_path, "w") as f:
            json.dump({"pad_token_id": self._pad_token_id}, f)

    @classmethod
    async def from_saved(
        cls: Type[BaseType], directory: Path, **kwargs
    ) -> BaseType:
        model_directory = directory / "model"
        config_path = directory / "config.json"

        with open(config_path, "r") as f:
            config = json.load(f)

        model = await background(cls._load_saved_model, model_directory)
        return cls(model, config["pad_token_id"])

    @classmethod
    @abstractmethod
    def _load_saved_model(cls, directory: Path) -> PreTrainedModel:
        pass

    @property
    def base_model(self) -> nn.Module:
        return self._model.base_model

    def freeze(self, pattern: str = ".*") -> None:
        pattern = re.compile(pattern)

        for name, parameter in self._model.base_model.named_parameters():
            if pattern.match(name):
                parameter.requires_grad = False


class HuggingfaceLanguageModel(HuggingfaceSequentialModelBase):
    @classmethod
    def _load_saved_model(cls, directory: Path) -> PreTrainedModel:
        return AutoModelForCausalLM.from_pretrained(directory)

    @classmethod
    async def from_path(cls: Type[BaseType], path: str) -> BaseType:
        model = await background(AutoModelForCausalLM.from_pretrained, path)
        if model.config.pad_token_id is not None:
            pad_token_id = model.config.pad_token_id
        elif model.config.eos_token_id is not None:
            pad_token_id = model.config.eos_token_id
        else:
            pad_token_id = 0
        return cls(model, pad_token_id)

    def freeze(self, pattern: str = ".*") -> None:
        super().freeze(pattern)

        for parameter in self._model.lm_head.parameters():
            parameter.requires_grad = True

    def forward(self, x: PackedSequence) -> PackedSequence:
        x, lengths = unpack_to_padded(x, pad_value=self._pad_token_id)
        x = x[:, :, 0]
        mask = make_mask(x, lengths)
        y = self._model(x, attention_mask=mask)
        return pack_padded(y.logits.log_softmax(-1), lengths)


class HuggingfaceValueModel(HuggingfaceSequentialModelBase):
    @classmethod
    def _load_saved_model(cls, directory: Path) -> PreTrainedModel:
        return AutoModelForTokenClassification.from_pretrained(directory)

    @classmethod
    async def from_path(cls: Type[BaseType], path: str) -> BaseType:
        config = await background(AutoConfig.from_pretrained, path)
        if config.pad_token_id is not None:
            pad_token_id = config.pad_token_id
        elif config.eos_token_id is not None:
            pad_token_id = config.eos_token_id
        else:
            pad_token_id = 0
        with suppress_huggingface_logging():
            model = await background(
                AutoModelForTokenClassification.from_pretrained,
                path,
                num_labels=1,
                problem_type="regression",
                pad_token_id=pad_token_id,
                ignore_mismatched_sizes=True,
            )
        return cls(model, pad_token_id)

    def freeze(self, pattern: str = ".*") -> None:
        super().freeze(pattern)

        for parameter in self._model.classifier.parameters():
            parameter.requires_grad = True

    def forward(self, x: PackedSequence) -> PackedSequence:
        x, lengths = unpack_to_padded(x, pad_value=self._pad_token_id)
        x = x[:, :, 0]
        mask = make_mask(x, lengths)
        y = self._model(x, attention_mask=mask)
        return pack_padded(y.logits, lengths)



from asyncio import Lock
from pathlib import Path
from typing import Type, Optional

from huggingface_hub.utils import HFValidationError

from kilroy_module_huggingface.models.models import (
    HuggingfaceSequentialModelBase,
)
from kilroy_module_huggingface.models.tokenizer import HuggingfaceTokenizer
from kilroy_module_pytorch_py_sdk import ModelLoader, SequentialModel
from kilroy_module_pytorch_py_sdk.models.loader import ModelInfo, ModelType
from kilroy_server_py_utils import Savable


class HuggingfaceModelLoader(ModelLoader[SequentialModel]):
    def __init__(
        self,
        lock: Lock,
        state_directory: Path,
        model_class: Type[HuggingfaceSequentialModelBase],
        path: str,
        batch_size: int,
        freeze: Optional[str] = None,
    ) -> None:
        self._state_directory = state_directory
        self._model_class = model_class
        self._path = path
        self._batch_size = batch_size
        self._freeze = freeze
        super().__init__(lock)

    @property
    def _model_directory(self) -> Path:
        return self._state_directory / "model"

    @property
    def _tokenizer_directory(self) -> Path:
        return self._state_directory / "tokenizer"

    async def _load(self) -> ModelInfo[SequentialModel]:
        try:
            model = await self._model_class.from_saved(self._model_directory)
        except (FileNotFoundError, HFValidationError):
            model = await self._model_class.from_path(self._path)

        try:
            tokenizer = await HuggingfaceTokenizer.from_saved(
                self._tokenizer_directory
            )
        except (FileNotFoundError, HFValidationError):
            tokenizer = await HuggingfaceTokenizer.from_path(self._path)

        if self._freeze is not None:
            model.freeze(self._freeze)

        return ModelInfo(model, tokenizer, self._batch_size, Lock())

    async def _save(self, info: ModelInfo[SequentialModel]) -> None:
        if isinstance(info.model, Savable):
            await info.model.save(self._model_directory)
        if isinstance(info.tokenizer, Savable):
            await info.tokenizer.save(self._tokenizer_directory)

    async def _reset(self) -> ModelInfo[ModelType]:
        model = await self._model_class.from_path(self._path)
        tokenizer = await HuggingfaceTokenizer.from_path(self._path)

        if self._freeze is not None:
            model.freeze(self._freeze)

        return ModelInfo(model, tokenizer, self._batch_size, Lock())



from contextlib import contextmanager

from transformers import logging


@contextmanager
def suppress_huggingface_logging(level=logging.ERROR):
    current_level = logging.get_verbosity()
    logging.set_verbosity(level)
    yield
    logging.set_verbosity(current_level)



import logging
from pathlib import Path
from typing import AsyncIterable, Dict, Any, Tuple, Optional

from kilroy_module_huggingface.models.loader import HuggingfaceModelLoader
from kilroy_module_huggingface.models.models import (
    HuggingfaceLanguageModel,
    HuggingfaceValueModel,
)
from kilroy_module_py_shared import Metadata, SerializableModel
from kilroy_module_pytorch_py_sdk import (
    PytorchModule,
    ModelsRegistry,
    ModelLoader,
    SequentialModel,
)
from kilroy_module_pytorch_py_sdk.module.params import ModelsParams
from kilroy_server_py_utils import classproperty, Configuration
from kilroy_server_py_utils.configurable import StateType


class LoaderParams(SerializableModel):
    path: str
    batch_size: int = 1
    freeze: Optional[str] = None


class HuggingfaceModule(PytorchModule):
    def __init__(
        self, config: Configuration[StateType], state_directory: Path, **kwargs
    ) -> None:
        self._state_directory = state_directory
        super().__init__(config, **kwargs)

    @staticmethod
    async def _build_policy_loader(
        state_directory: Path, params: Dict[str, Any]
    ) -> ModelLoader[SequentialModel]:
        params = LoaderParams(**params)
        return await HuggingfaceModelLoader.create(
            state_directory=state_directory,
            model_class=HuggingfaceLanguageModel,
            path=params.path,
            batch_size=params.batch_size,
            freeze=params.freeze,
        )

    @staticmethod
    async def _build_value_loader(
        state_directory: Path, params: Dict[str, Any]
    ) -> ModelLoader[SequentialModel]:
        params = LoaderParams(**params)
        return await HuggingfaceModelLoader.create(
            state_directory=state_directory,
            model_class=HuggingfaceValueModel,
            path=params.path,
            batch_size=params.batch_size,
            freeze=params.freeze,
        )

    @staticmethod
    async def _build_baseline_loader(
        state_directory: Path, policy_params: Dict[str, Any]
    ) -> ModelLoader[SequentialModel]:
        params = LoaderParams(**policy_params)
        return await HuggingfaceModelLoader.create(
            state_directory=state_directory,
            model_class=HuggingfaceLanguageModel,
            path=params.path,
            batch_size=params.batch_size,
            freeze=".*",
        )

    async def _build_models_registry(
        self, params: ModelsParams
    ) -> ModelsRegistry:
        state_directory = self._state_directory / "models"
        return ModelsRegistry(
            policy=await self._build_policy_loader(
                state_directory / "policy", params.policy
            ),
            value=await self._build_value_loader(
                state_directory / "value", params.value
            ),
            baseline=await self._build_baseline_loader(
                state_directory / "baseline", params.policy
            ),
        )

    # noinspection PyMethodParameters
    @classproperty
    def metadata(cls) -> Metadata:
        return Metadata(
            key="kilroy-module-huggingface",
            description="Kilroy module for Huggingface models",
        )

    # noinspection PyMethodParameters
    @classproperty
    def logger(cls) -> logging.Logger:
        return logging.getLogger(__name__)

    async def generate(
        self, n: int
    ) -> AsyncIterable[Tuple[Dict[str, Any], Dict[str, Any]]]:
        self.logger.info(f"Generating {n} posts...")

        i = 0

        async for content, metadata in super().generate(n):
            self.logger.info(f"Generated post {i}.")
            yield content, metadata
            i += 1

        self.logger.info("Finished generating posts.")

    async def fit_supervised(
        self, data: AsyncIterable[Tuple[Dict[str, Any], float]]
    ) -> None:
        self.logger.info("Fitting supervised...")
        await super().fit_supervised(data)
        self.logger.info(f"Finished fitting supervised.")

    async def fit_reinforced(
        self, data: AsyncIterable[Tuple[Dict[str, Any], Dict[str, Any], float]]
    ) -> None:
        self.logger.info(f"Fitting reinforced...")
        await super().fit_reinforced(data)
        self.logger.info(f"Finished fitting reinforced.")

    async def reset_self(self) -> None:
        self.logger.info("Resetting state...")
        await super().reset_self()
        self.logger.info("State reset.")

    async def save_self(self, directory: Path) -> None:
        self.logger.info("Saving state...")
        await super().save_self(directory)
        self.logger.info("State saved.")
